@@ -9,20 +9,18 @@ from datetime import datetime
 from functools import partial
 import logging
 import random
-from typing import TYPE_CHECKING, Any, Callable, Iterable, TypeVar, cast
+from typing import TYPE_CHECKING, Any, Awaitable, Callable, Iterable, TypeVar, cast
 
-from chip.ChipDeviceCtrl import (
-    CommissionableNode,
-    CommissioningParameters,
-    DeviceProxyWrapper,
-)
+from chip.ChipDeviceCtrl import DeviceProxyWrapper
 from chip.clusters import Attribute, Objects as Clusters
 from chip.clusters.Attribute import ValueDecodeFailure
 from chip.clusters.ClusterObjects import ALL_ATTRIBUTES, ALL_CLUSTERS, Cluster
-from chip.discovery import CommissionableNode as CommissionableNodeData
 from chip.exceptions import ChipStackError
 
+from matter_server.common.helpers.util import convert_ip_address
+from matter_server.common.models import CommissionableNodeData, CommissioningParameters
 from matter_server.server.helpers.attributes import parse_attributes_from_read_result
+from matter_server.server.helpers.utils import ping_ip
 
 from ..common.const import SCHEMA_VERSION
 from ..common.errors import (
@@ -35,9 +33,17 @@ from ..common.helpers.api import api_command
 from ..common.helpers.util import (
     create_attribute_path_from_attribute,
     dataclass_from_dict,
+    dataclass_to_dict,
     parse_attribute_path,
+    parse_value,
+)
+from ..common.models import (
+    APICommand,
+    EventType,
+    MatterNodeData,
+    MatterNodeEvent,
+    NodePingResult,
 )
-from ..common.models import APICommand, EventType, MatterNodeData, MatterNodeEvent
 from .const import PAA_ROOT_CERTS_DIR
 from .helpers.paa_certificates import fetch_certificates
 
@@ -55,7 +61,11 @@ LOGGER = logging.getLogger(__name__)
 MAX_POLL_INTERVAL = 600
 MAX_COMMISSION_RETRIES = 3
 
-BASE_SUBSCRIBE_ATTRIBUTES: tuple[Attribute.AttributePath, Attribute.AttributePath] = (
+ROUTING_ROLE_ATTRIBUTE_PATH = create_attribute_path_from_attribute(
+    0, Clusters.ThreadNetworkDiagnostics.Attributes.RoutingRole
+)
+
+BASE_SUBSCRIBE_ATTRIBUTES: tuple[Attribute.AttributePath, ...] = (
     # all endpoints, BasicInformation cluster
     Attribute.AttributePath(
         EndpointId=None, ClusterId=Clusters.BasicInformation.id, Attribute=None
@@ -66,6 +76,15 @@ BASE_SUBSCRIBE_ATTRIBUTES: tuple[Attribute.AttributePath, Attribute.AttributePat
         ClusterId=Clusters.BridgedDeviceBasicInformation.id,
         Attribute=None,
     ),
+    # networkinterfaces attribute on general diagnostics cluster,
+    # so we have the most accurate IP addresses for ping/diagnostics
+    Attribute.AttributePath(
+        EndpointId=0, Attribute=Clusters.GeneralDiagnostics.Attributes.NetworkInterfaces
+    ),
+    # active fabrics attribute - to speedup node diagnostics
+    Attribute.AttributePath(
+        EndpointId=0, Attribute=Clusters.OperationalCredentials.Attributes.Fabrics
+    ),
 )
 
 # pylint: disable=too-many-lines,too-many-locals,too-many-statements,too-many-branches
@@ -88,12 +107,13 @@ class MatterDeviceController:
         self._attr_subscriptions: dict[int, list[Attribute.AttributePath]] = {}
         self._resub_debounce_timer: dict[int, asyncio.TimerHandle] = {}
         self._sub_retry_timer: dict[int, asyncio.TimerHandle] = {}
-        self._nodes: dict[int, MatterNodeData | None] = {}
+        self._nodes: dict[int, MatterNodeData] = {}
         self._last_subscription_attempt: dict[int, int] = {}
         self.wifi_credentials_set: bool = False
         self.thread_credentials_set: bool = False
         self.compressed_fabric_id: int | None = None
         self._node_lock: dict[int, asyncio.Lock] = {}
+        self._resolve_lock = asyncio.Lock()
 
     async def initialize(self) -> None:
         """Async initialize of controller."""
@@ -113,20 +133,22 @@ class MatterDeviceController:
         """Handle logic on controller start."""
         # load nodes from persistent storage
         nodes: dict[str, dict | None] = self.server.storage.get(DATA_KEY_NODES, {})
+        orphaned_nodes: set[str] = set()
         for node_id_str, node_dict in nodes.items():
             node_id = int(node_id_str)
             if node_dict is None:
-                # ignore non-initialized (left-over) nodes
-                # from failed commissioning attempts
+                # Non-initialized (left-over) node from a failed commissioning attempt.
+                # NOTE: This code can be removed in a future version
+                # as this can no longer happen.
+                orphaned_nodes.add(node_id_str)
                 continue
             if node_dict.get("interview_version") != SCHEMA_VERSION:
-                # invalidate node data if schema mismatch,
-                # the node will automatically be scheduled for re-interview
-                node = None
-            else:
-                node = dataclass_from_dict(MatterNodeData, node_dict)
-                # always mark node as unavailable at startup until subscriptions are ready
-                node.available = False
+                # Invalidate node attributes data if schema mismatch,
+                # the node will automatically be scheduled for re-interview.
+                node_dict["attributes"] = {}
+            node = dataclass_from_dict(MatterNodeData, node_dict)
+            # always mark node as unavailable at startup until subscriptions are ready
+            node.available = False
             self._nodes[node_id] = node
             # setup subscription and (re)interview as task in the background
             # as we do not want it to block our startup
@@ -135,9 +157,12 @@ class MatterDeviceController:
                 # the first attempt to initialize so that we prioritize nodes
                 # that are probably available so they are back online as soon as
                 # possible and we're not stuck trying to initialize nodes that are offline
-                self._schedule_interview(node_id, 5)
+                self._schedule_interview(node_id, 30)
             else:
                 asyncio.create_task(self._check_interview_and_subscription(node_id))
+        # cleanup orhpaned nodes from storage
+        for node_id_str in orphaned_nodes:
+            self.server.storage.remove(DATA_KEY_NODES, node_id_str)
         LOGGER.info("Loaded %s nodes from stored configuration", len(self._nodes))
 
     async def stop(self) -> None:
@@ -374,7 +399,7 @@ class MatterDeviceController:
         if discriminator is None:
             discriminator = 3840  # TODO generate random one
 
-        return await self._call_sdk(
+        sdk_result = await self._call_sdk(
             self.chip_controller.OpenCommissioningWindow,
             nodeid=node_id,
             timeout=timeout,
@@ -382,29 +407,49 @@ class MatterDeviceController:
             discriminator=discriminator,
             option=option,
         )
+        return CommissioningParameters(
+            setup_pin_code=sdk_result.setupPinCode,
+            setup_manual_code=sdk_result.setupManualCode,
+            setup_qr_code=sdk_result.setupQRCode,
+        )
 
     @api_command(APICommand.DISCOVER)
     async def discover_commissionable_nodes(
         self,
-    ) -> CommissionableNodeData | list[CommissionableNodeData] | None:
+    ) -> list[CommissionableNodeData]:
         """Discover Commissionable Nodes (discovered on BLE or mDNS)."""
         if self.chip_controller is None:
             raise RuntimeError("Device Controller not initialized.")
 
-        result = await self._call_sdk(
+        sdk_result = await self._call_sdk(
             self.chip_controller.DiscoverCommissionableNodes,
         )
-
-        def convert(cn: CommissionableNode) -> CommissionableNodeData:
-            cnd = CommissionableNodeData()
-            # pylint: disable=no-member
-            for field in CommissionableNodeData.__dataclass_fields__:
-                setattr(cnd, field, getattr(cn, field))
-            return cnd
-
-        if isinstance(result, list):
-            return [convert(c) for c in result]
-        return convert(result)
+        if sdk_result is None:
+            return []
+        # ensure list
+        if not isinstance(sdk_result, list):
+            sdk_result = [sdk_result]
+        return [
+            CommissionableNodeData(
+                instance_name=x.instanceName,
+                host_name=x.hostName,
+                port=x.port,
+                long_discriminator=x.longDiscriminator,
+                vendor_id=x.vendorId,
+                product_id=x.productId,
+                commissioning_mode=x.commissioningMode,
+                device_type=x.deviceType,
+                device_name=x.deviceName,
+                pairing_instruction=x.pairingInstruction,
+                pairing_hint=x.pairingHint,
+                mrp_retry_interval_idle=x.mrpRetryIntervalIdle,
+                mrp_retry_interval_active=x.mrpRetryIntervalActive,
+                supports_tcp=x.supportsTcp,
+                addresses=x.addresses,
+                rotating_id=x.rotatingId,
+            )
+            for x in sdk_result
+        ]
 
     @api_command(APICommand.INTERVIEW_NODE)
     async def interview_node(self, node_id: int) -> None:
@@ -448,12 +493,7 @@ class MatterDeviceController:
 
         # save updated node data
         self._nodes[node_id] = node
-        self.server.storage.set(
-            DATA_KEY_NODES,
-            subkey=str(node_id),
-            value=node,
-            force=not existing_info,
-        )
+        self._write_node_state(node_id, True)
         if is_new_node:
             # new node - first interview
             self.server.signal_event(EventType.NODE_ADDED, node)
@@ -483,7 +523,8 @@ class MatterDeviceController:
         cluster_cls: Cluster = ALL_CLUSTERS[cluster_id]
         command_cls = getattr(cluster_cls.Commands, command_name)
         command = dataclass_from_dict(command_cls, payload)
-        async with self._get_node_lock(node_id):
+        node_lock = self._get_node_lock(node_id)
+        async with node_lock:
             return await self.chip_controller.SendCommand(
                 nodeid=node_id,
                 endpoint=endpoint_id,
@@ -495,19 +536,17 @@ class MatterDeviceController:
 
     @api_command(APICommand.READ_ATTRIBUTE)
     async def read_attribute(
-        self,
-        node_id: int,
-        attribute_path: str,
+        self, node_id: int, attribute_path: str, fabric_filtered: bool = False
     ) -> Any:
         """Read a single attribute (or Cluster) on a node."""
         if self.chip_controller is None:
             raise RuntimeError("Device Controller not initialized.")
-        node_lock = self._get_node_lock(node_id)
         endpoint_id, cluster_id, attribute_id = parse_attribute_path(attribute_path)
+        assert self.server.loop is not None
+        future = self.server.loop.create_future()
+        device = await self._resolve_node(node_id)
+        node_lock = self._get_node_lock(node_id)
         async with node_lock:
-            assert self.server.loop is not None
-            future = self.server.loop.create_future()
-            device = await self._resolve_node(node_id)
             Attribute.Read(
                 future=future,
                 eventLoop=self.server.loop,
@@ -520,13 +559,13 @@ class MatterDeviceController:
                         AttributeId=attribute_id,
                     )
                 ],
+                fabricFiltered=fabric_filtered,
             ).raise_on_error()
             result: Attribute.AsyncReadTransaction.ReadResponse = await future
             read_atributes = parse_attributes_from_read_result(result.tlvAttributes)
             # update cached info in node attributes
-            self._nodes[node_id].attributes.update(  # type: ignore[union-attr]
-                read_atributes
-            )
+            self._nodes[node_id].attributes.update(read_atributes)
+            self._write_node_state(node_id)
             if len(read_atributes) > 1:
                 return read_atributes
             return read_atributes.get(attribute_path, None)
@@ -575,7 +614,6 @@ class MatterDeviceController:
             DATA_KEY_NODES,
             subkey=str(node_id),
         )
-        self.server.storage.save(immediate=True)
         LOGGER.info("Node ID %s successfully removed from Matter server.", node_id)
 
         self.server.signal_event(EventType.NODE_REMOVED, node_id)
@@ -642,11 +680,7 @@ class MatterDeviceController:
         if prev_subs == node.attribute_subscriptions:
             return  # nothing to do
         # save updated node data
-        self.server.storage.set(
-            DATA_KEY_NODES,
-            subkey=str(node_id),
-            value=node,
-        )
+        self._write_node_state(node_id)
 
         # (re)setup node subscription
         # this could potentially be called multiple times within a short timeframe
@@ -662,6 +696,62 @@ class MatterDeviceController:
             5, resubscribe
         )
 
+    @api_command(APICommand.PING_NODE)
+    async def ping_node(self, node_id: int) -> NodePingResult:
+        """Ping node on the currently known IP-adress(es)."""
+        result: NodePingResult = {}
+        # the node's ip addresses are stored in the GeneralDiagnostics cluster
+        attribute = Clusters.GeneralDiagnostics.Attributes.NetworkInterfaces
+        attr_path = f"0/{attribute.cluster_id}/{attribute.attribute_id}"
+        node = self._nodes.get(node_id)
+        if node is None:
+            raise NodeNotExists(
+                f"Node {node_id} does not exist or is not yet interviewed"
+            )
+
+        battery_powered = (
+            node.attributes.get(ROUTING_ROLE_ATTRIBUTE_PATH, 0)
+            == Clusters.ThreadNetworkDiagnostics.Enums.RoutingRoleEnum.kSleepyEndDevice
+        )
+
+        async def _do_ping(ip_address: str) -> None:
+            """Ping IP and add to result."""
+            timeout = 10 if battery_powered else 2
+            result[ip_address] = await ping_ip(ip_address, timeout)
+
+        # The network interfaces attribute contains a list of network interfaces.
+        # For regular nodes this is just a single interface but we iterate them all anyway.
+        # Create a list of tasks so we can do multiple pings simultanuous.
+        # NOTE: Upgrade this to a TaskGroup once we bump our minimal python version.
+        attr_data = cast(list[dict[str, Any]], node.attributes.get(attr_path))
+        tasks: list[Awaitable] = []
+        for network_interface_data in attr_data:
+            network_interface: Clusters.GeneralDiagnostics.Structs.NetworkInterface = (
+                parse_value(
+                    "network_interface",
+                    network_interface_data,
+                    Clusters.GeneralDiagnostics.Structs.NetworkInterface,
+                )
+            )
+            # ignore invalid/non-operational interfaces
+            if not network_interface.isOperational:
+                continue
+            if network_interface.type in (
+                Clusters.GeneralDiagnostics.Enums.InterfaceTypeEnum.kUnspecified,
+                Clusters.GeneralDiagnostics.Enums.InterfaceTypeEnum.kUnknownEnumValue,
+            ):
+                continue
+
+            # enumerate ipv4 and ipv6 addresses
+            for ipv4_address_hex in network_interface.IPv4Addresses:
+                ipv4_address = convert_ip_address(ipv4_address_hex)
+                tasks.append(_do_ping(ipv4_address))
+            for ipv6_address_hex in network_interface.IPv6Addresses:
+                ipv6_address = convert_ip_address(ipv6_address_hex, True)
+                tasks.append(_do_ping(ipv6_address))
+        await asyncio.gather(*tasks)
+        return result
+
     async def _subscribe_node(self, node_id: int) -> None:
         """
         Subscribe to all node state changes/events for an individual node.
@@ -680,7 +770,7 @@ class MatterDeviceController:
 
         node_logger = LOGGER.getChild(f"[node {node_id}]")
         node_lock = self._get_node_lock(node_id)
-        node = cast(MatterNodeData, self._nodes[node_id])
+        node = self._nodes[node_id]
 
         # work out all (current) attribute subscriptions
         attr_subscriptions: list[Attribute.AttributePath] = list(
@@ -728,7 +818,7 @@ class MatterDeviceController:
         # determine if node is battery powered sleeping device
         # Endpoint 0, ThreadNetworkDiagnostics Cluster, routingRole attribute
         battery_powered = (
-            node.attributes.get("0/53/1", 0)
+            node.attributes.get(ROUTING_ROLE_ATTRIBUTE_PATH, 0)
             == Clusters.ThreadNetworkDiagnostics.Enums.RoutingRoleEnum.kSleepyEndDevice
         )
 
@@ -750,6 +840,10 @@ class MatterDeviceController:
             attr_path = str(path.Path)
             old_value = node.attributes.get(attr_path)
 
+            # return early if the value did not actually change at all
+            if old_value == new_value:
+                return
+
             node_logger.debug(
                 "Attribute updated: %s - old value: %s - new value: %s",
                 path,
@@ -787,11 +881,7 @@ class MatterDeviceController:
             node.attributes[attr_path] = new_value
 
             # schedule save to persistent storage
-            self.server.storage.set(
-                DATA_KEY_NODES,
-                subkey=str(node_id),
-                value=node,
-            )
+            self._write_node_state(node_id)
 
             # This callback is running in the CHIP stack thread
             loop.call_soon_threadsafe(
@@ -862,15 +952,15 @@ class MatterDeviceController:
                 node.available = True
                 self.server.signal_event(EventType.NODE_UPDATED, node)
 
+        node_logger.info("Setting up attributes and events subscription.")
+        interval_floor = 0
+        interval_ceiling = (
+            random.randint(60, 300) if battery_powered else random.randint(30, 120)
+        )
+        self._last_subscription_attempt[node_id] = 0
+        future = loop.create_future()
+        device = await self._resolve_node(node_id)
         async with node_lock:
-            node_logger.info("Setting up attributes and events subscription.")
-            interval_floor = 0
-            interval_ceiling = (
-                random.randint(60, 300) if battery_powered else random.randint(30, 120)
-            )
-            self._last_subscription_attempt[node_id] = 0
-            future = loop.create_future()
-            device = await self._resolve_node(node_id)
             Attribute.Read(
                 future=future,
                 eventLoop=loop,
@@ -891,13 +981,13 @@ class MatterDeviceController:
                 fabricFiltered=False,
                 autoResubscribe=True,
             ).raise_on_error()
-            sub: Attribute.SubscriptionTransaction = await future
+        sub: Attribute.SubscriptionTransaction = await future
 
-            sub.SetAttributeUpdateCallback(attribute_updated_callback)
-            sub.SetEventUpdateCallback(event_callback)
-            sub.SetErrorCallback(error_callback)
-            sub.SetResubscriptionAttemptedCallback(resubscription_attempted)
-            sub.SetResubscriptionSucceededCallback(resubscription_succeeded)
+        sub.SetAttributeUpdateCallback(attribute_updated_callback)
+        sub.SetEventUpdateCallback(event_callback)
+        sub.SetErrorCallback(error_callback)
+        sub.SetResubscriptionAttemptedCallback(resubscription_attempted)
+        sub.SetResubscriptionSucceededCallback(resubscription_succeeded)
 
         # if we reach this point, it means the node could be resolved
         # and the initial subscription succeeded, mark the node available.
@@ -1015,12 +1105,13 @@ class MatterDeviceController:
                 attempt,
                 retries,
             )
-            return await self._call_sdk(
-                self.chip_controller.GetConnectedDeviceSync,
-                nodeid=node_id,
-                allowPASE=False,
-                timeoutMs=None,
-            )
+            async with self._resolve_lock:
+                return await self._call_sdk(
+                    self.chip_controller.GetConnectedDeviceSync,
+                    nodeid=node_id,
+                    allowPASE=False,
+                    timeoutMs=None,
+                )
         except (ChipStackError, TimeoutError) as err:
             if attempt >= retries:
                 # when we're out of retries, raise NodeNotResolving
@@ -1033,7 +1124,7 @@ class MatterDeviceController:
 
     def _handle_endpoints_removed(self, node_id: int, endpoints: Iterable[int]) -> None:
         """Handle callback for when bridge endpoint(s) get deleted."""
-        node = cast(MatterNodeData, self._nodes[node_id])
+        node = self._nodes[node_id]
         for endpoint_id in endpoints:
             node.attributes = {
                 key: value
@@ -1045,11 +1136,7 @@ class MatterDeviceController:
                 {"node_id": node_id, "endpoint_id": endpoint_id},
             )
         # schedule save to persistent storage
-        self.server.storage.set(
-            DATA_KEY_NODES,
-            subkey=str(node_id),
-            value=node,
-        )
+        self._write_node_state(node_id)
 
     async def _handle_endpoints_added(
         self, node_id: int, endpoints: Iterable[int]
@@ -1069,3 +1156,13 @@ class MatterDeviceController:
         if node_id not in self._node_lock:
             self._node_lock[node_id] = asyncio.Lock()
         return self._node_lock[node_id]
+
+    def _write_node_state(self, node_id: int, force: bool = False) -> None:
+        """Schedule the write of the current node state to persistent storage."""
+        node = self._nodes[node_id]
+        self.server.storage.set(
+            DATA_KEY_NODES,
+            value=dataclass_to_dict(node),
+            subkey=str(node_id),
+            force=force,
+        )
